/*!
 * \author Ruben Martins - ruben@sat.inesc-id.pt
 *
 * @section LICENSE
 *
 * Open-WBO, Copyright (c) 2013-2017, Ruben Martins, Vasco Manquinho, Ines Lynce
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef Alg_OLL_h
#define Alg_OLL_h

#ifdef SIMP
#include "simp/SimpSolver.h"
#else
#include "core/Solver.h"
#endif

#include "../Encoder.h"
#include "../MaxSAT.h"
#include <map>
#include <set>

namespace openwbo {

//=================================================================================================
class OLL : public MaxSAT {

public:
  OLL(int verb = _VERBOSITY_MINIMAL_, bool preprocessor = false) {
    solver = NULL;
    verbosity = verb;
    incremental_strategy = _INCREMENTAL_ITERATIVE_;
    encoding = _CARD_TOTALIZER_;
    encoder.setCardEncoding(_CARD_TOTALIZER_);
    min_weight = 1;
    bestCost = UINT64_MAX;
    _unitCores = true;
    _amo = true;
    _preprocessor = preprocessor;
  }
  ~OLL() {
    if (solver != NULL)
      delete solver;
  }

  StatusCode search();

  // Print solver configuration.
  void printConfiguration() {

    if(!print) return;

    printf("c ==========================================[ Solver Settings "
           "]============================================\n");
    printf("c |                                                                "
           "                                       |\n");
    printf("c |  Algorithm: %23s                                             "
           "                      |\n",
           "OLL");
    print_Card_configuration(encoding);
    printf("c |                                                                "
           "                                       |\n");
  }

// refactor this code and create an interface on Utils
  std::vector<std::vector<Lit>> unit_cores;
  std::vector<std::vector<Lit>> am1;

  void preprocessor();

protected:
  // Rebuild MaxSAT solver
  //
  Solver *rebuildSolver(); // Rebuild MaxSAT solver.

  // Other
  void initRelaxation(); // Relaxes soft clauses.

  StatusCode unweighted();
  StatusCode weighted();

  Solver *solver;  // SAT Solver used as a black box.
  Encoder encoder; // Interface for the encoder of constraints to CNF.

  // Controls the incremental strategy used by MSU3 algorithms.
  int incremental_strategy;
  // Controls the cardinality encoding used by MSU3 algorithms.
  int encoding;

  // Literals to be used in the constraint that excludes models.
  vec<Lit> objFunction;
  vec<uint64_t> coeffs; // Coefficients of the literals that are used in the
                        // constraint that excludes models.

  std::map<Lit, int> coreMapping; // Mapping between the assumption literal and
                                  // the respective soft clause.

  std::map<Lit, int>
      coreCardinality; // Mapping between the assumption literal and
                       // the respective soft clause.

  // lit -> <ID, bound, weight>
  std::map<Lit, std::pair<std::pair<int, uint64_t>, uint64_t>> boundMapping;

  // Soft clauses that are currently in the MaxSAT formula.
  vec<bool> activeSoft;

  uint64_t findNextWeightDiversity(uint64_t weight,
                                   std::set<Lit> &cardinality_assumptions);
  uint64_t findNextWeight(uint64_t weight,
                          std::set<Lit> &cardinality_assumptions);

  uint64_t min_weight;

  // improvements
  int findUPUnitCores(Solver *sat_solver);
  uint64_t bestCost;
  uint64_t findAtMost1(Solver * sat_solver);
  void bronKerbosch(std::vector<Lit> R, std::vector<Lit> P, std::vector<Lit> X);
  std::map<Lit, std::vector<Lit>> graph;
  std::map<Lit, bool> graph_used;
  uint64_t lower_bound;
  std::vector<std::vector<Lit>> disjoint_cores;
  std::vector<std::vector<Lit>> amo_cores;
  int itn_current;


  bool _preprocessor;

static std::vector<Lit> vintersection(std::vector<Lit> &v1, std::vector<Lit> &v2) {
  std::vector<Lit> v;

  sort(v1.begin(), v1.end());
  sort(v2.begin(), v2.end());

  set_intersection(v1.begin(),v1.end(),v2.begin(),v2.end(),back_inserter(v));

  return v;
}

static std::vector<Lit> vunion(std::vector<Lit> &v1, std::vector<Lit> &v2) {
  std::vector<Lit> v;

  sort(v1.begin(), v1.end());
  sort(v2.begin(), v2.end());

  set_union(v1.begin(),v1.end(),v2.begin(),v2.end(),back_inserter(v));

  return v;
}

static std::vector<Lit> vdifference(std::vector<Lit> &v1, std::vector<Lit> &v2) {
  std::vector<Lit> v;

  sort(v1.begin(), v1.end());
  sort(v2.begin(), v2.end());

  set_difference(v1.begin(),v1.end(),v2.begin(),v2.end(),back_inserter(v));

  return v;
}

void findDisjointCores(Solver * sat_solver);

  bool _unitCores;
  bool _disjointCores;
  bool _amo;

  

};
} // namespace openwbo

#endif
